import logging

from allauth.account.models import EmailAddress
from django.contrib.auth import authenticate
from django.contrib.auth import get_user_model
from django.db import IntegrityError
from django.forms import ValidationError
from django.http import Http404
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.cache import cache_page
from rest_framework import generics
from rest_framework import mixins
from rest_framework import status
from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.views import APIView

from django_sso_app.backend.utils import invalidate_cookie, get_request_jwt, update_response_jwt
from .serializers import (CheckUserExistenceSerializer,
                          UserSerializer, UserUpdateSerializer,
                          UserUnsubscriptionSerializer, UserRevisionSerializer)
from ...permissions import StaffPermission
from ...permissions import is_staff

logger = logging.getLogger('users')
User = get_user_model()


class TryAuthenticateMixin(object):

    def try_authenticate(self, username, email, password):
        credentials = username or email or None
        user = None

        if username is None and email is not None:
            credentials = 'email'
        if username is not None and email is None:
            credentials = 'username'

        if credentials is not None:
            user = authenticate(username=credentials, password=password)

        if user is not None:
            return user

        raise User.DoesNotExist('Check credentials.')


class CheckUserExistence(APIView):
    """
    Retrieve a "condensed" user instance {"id": ".."} by either "username" or "email" as query params.
    """

    def get_object(self, username=None, email=None):
        try:
            if username is None and email is not None:
                email_address = EmailAddress.objects.get(email=email)
                return email_address.user
            if username is not None and email is None:
                return User.objects.get(username=username)
            raise User.DoesNotExist('Either username or email must be set.')
        except (User.DoesNotExist, EmailAddress.DoesNotExist):
            raise Http404

    def get(self, *args, **kwargs):
        username = self.request.query_params.get('username', None)
        email = self.request.query_params.get('email', None)

        if username is not None or email is not None:
            item = self.get_object(username, email)
        else:
            raise Http404

        serializer = CheckUserExistenceSerializer(item)
        return Response(serializer.data)


class UserApiViewSet(mixins.ListModelMixin,
                     mixins.UpdateModelMixin,
                     mixins.RetrieveModelMixin,
                     viewsets.GenericViewSet):
    """
    User api viewset
    """

    lookup_field = 'sso_id'

    def get_queryset(self):
        user = self.request.user
        if is_staff(user):
            queryset = User.objects.all()
        else:
            queryset = User.objects.filter(sso_id=user.sso_id, is_staff=False)
        return queryset

    def get_serializer_class(self, *args, **kwargs):
        if self.request.method in ['PATCH']:
            return UserUpdateSerializer
        else:
            return UserSerializer

    def get_object(self, *args, **kwargs):
        pk = self.kwargs.get('pk')
        try:
            user = User.objects.get(sso_id=pk)
            self.check_object_permissions(self.request, user)
            return user
        except User.DoesNotExist:
            raise Http404

    def list(self, request, *args, **kwargs):
        """
        List users
        """
        return super(UserApiViewSet, self).list(request, *args, **kwargs)

    def partial_update(self, request, *args, **kwargs):
        try:
            response = super(UserApiViewSet, self).partial_update(request, *args, **kwargs)
        except (IntegrityError, ValidationError) as e:
            logger.info('User {0} tried to update email with already registered one'.format(request.user))
            return Response(_('Email already registered.'), status=status.HTTP_409_CONFLICT)

        response_status_code = response.status_code
        logger.info('Partial_update status code is {}, {}'.format(response_status_code, request.data))

        if response_status_code == 200:
            received_jwt = get_request_jwt(request)

            user = self.get_object()
            requesting_user = self.request.user

            from_browser_as_same_user = received_jwt is not None and requesting_user == user
            logger.info('From browser as same user? {}'. format(from_browser_as_same_user))

            if from_browser_as_same_user: # From browser as same user
                email_changed = requesting_user.email != user.email

                if email_changed:
                    logger.info('User {0} must confirm email'.format(user))
                else:
                    logger.info('User {0} must NOT confirm email, updating JWT'.format(user))
                    update_response_jwt(received_jwt, user, request, response)

        return response


class UserUnsubscriptionApiView(TryAuthenticateMixin, APIView):
    """
    Completely unsubscribes requesting user
    """

    def post(self, request, *args, **kwargs):
        user = request.user
        serializer = UserUnsubscriptionSerializer(data=request.data, context={'request': request})

        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        username = user.username
        email = user.email
        password = serializer.data.get('password')

        has_unsubscribed = False
        user = self.try_authenticate(username, email, password)

        logger.info('User {0} wants to unregister from SSO'.format(user))

        if user == request.user:
            user.unsubscribe()
            has_unsubscribed = True
            user.refresh_from_db()
        else:
            logger.error('User {0} tried to unregister user {1} from SSO'.format(request.user, user))

        serializer = UserSerializer(user, context={'request': request})
        response = Response(serializer.data)

        if has_unsubscribed:
            invalidate_cookie(response)

        return response


class UserRevisionsApiView(generics.ListAPIView):
    """
    User revisions entrypoint
    """

    queryset = User.objects.all()
    serializer_class = UserRevisionSerializer
    permission_classes = (StaffPermission,)

    @method_decorator(cache_page(120))
    def list(self, request):
        # Note the use of `get_queryset()` instead of `self.queryset`
        queryset = self.get_queryset()
        serializer = self.serializer_class(queryset, many=True)
        return Response(serializer.data)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _
from django.conf import settings


class UsersConfig(AppConfig):
    name = "django_sso_app.backend.users"
    verbose_name = _("Users")

    def ready(self):
        try:
            import django_sso_app.backend.users.signals  # noqa F401
            if settings.DJANGO_SSO_BACKEND_ENABLED:
                import django_sso_app.core.apps.users.signals  # noqa F401
        except ImportError:
            pass



from django.apps import AppConfig


class UsersConfig(AppConfig):
    name = 'django_sso_app.core.apps.users'

    def __init__(self, *args, **kwargs):
        ret = super(UsersConfig, self).__init__(*args, **kwargs)
        print('eccolo', self.label)
        return ret

    def ready(self):
        try:
            import django_sso_app.core.apps.users.signals  # noqa F401
        except ImportError:
            pass



import logging

from django.conf import settings
from rest_framework import serializers
from rest_framework.reverse import reverse

from django_sso_app.core.serializers import AbsoluteUrlSerializer
from django_sso_app.core.utils import get_profile_model

logger = logging.getLogger('profiles')
Profile = get_profile_model()


class ProfileSerializer(AbsoluteUrlSerializer):
    user = serializers.SerializerMethodField(required=False)
    sso_id = serializers.SerializerMethodField(required=False)
    sso_rev = serializers.SerializerMethodField(required=False)

    class Meta:
        model = Profile
        read_only_fields = (
            'url',
            'created_at',
            'sso_id', 'sso_rev',
            'username', 'email',
            'user') + settings.DJANGO_SSO_REQUIRED_USER_FIELDS
        fields = read_only_fields + settings.DJANGO_SSO_REQUIRED_PROFILE_FIELDS

    def get_absolute_url(self, obj):
        if getattr(obj, 'id', None) is not None:
            request = self.context['request']
            reverse_url = reverse('profile:detail', args=[obj.sso_id])
            return request.build_absolute_uri(reverse_url)

    def get_user(self, instance):
        try:
            user = instance.user
            request = self.context['request']
            return request.build_absolute_uri(user.get_absolute_url())
        except:
            logger.warning('No profile.user url for {}.'.format(instance))

    def get_sso_id(self, instance):
        if settings.DJANGO_SSO_BACKEND_ENABLED:
            return instance.user.sso_id
        else:
            return instance.sso_id

    def get_sso_rev(self, instance):
        if settings.DJANGO_SSO_BACKEND_ENABLED:
            return instance.user.sso_rev
        else:
            return instance.sso_rev


class ProfilePublicSerializer(ProfileSerializer):

    class Meta:
        model = Profile
        read_only_fields = (
            'url',
            'sso_id',
            'created_at', 'username', 'picture',
            'country')
        fields = read_only_fields



import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _

from django_sso_app.core.common import ROLE_CHOICES

logger = logging.getLogger('services')


class Service(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    name = models.CharField(max_length=32, unique=True, null=True)
    url = models.CharField(max_length=255, unique=True)
    picture = models.TextField(_('picture'), null=True, blank=True)

    cookie_domain = models.CharField(max_length=255, null=True, blank=True)  # unused but useful
    redirect_wait = models.PositiveIntegerField(default=2000)

    subscription_required = models.BooleanField(default=False)
    is_public = models.BooleanField(default=True)

    role = models.SmallIntegerField(
        choices=ROLE_CHOICES, null=True, blank=True)

    def get_tos(self, language):
        logger.debug('Service {0}, gettings tos for language {1}'.format(self, language))
        return self.terms_of_service.filter(language=language).first()

    def __str__(self):
        return self.url


class TOS(models.Model):
    language = models.CharField(max_length=3)
    text = models.TextField()

    service = models.ForeignKey(Service, related_name="terms_of_service", on_delete=models.CASCADE)


class Subscription(models.Model):
    user = models.ForeignKey(
        "users.User",
        on_delete=models.CASCADE,
        related_name="subscriptions"
    )
    service = models.ForeignKey(
        Service,
        on_delete=models.CASCADE,
        related_name="subscriptions"
    )
    created_at = models.DateTimeField(auto_now_add=True)
    unsubscribed_at = models.DateTimeField(null=True, blank=True)

    @property
    def is_unsubscribed(self):
        return self.unsubscribed_at is not None

    def __str__(self):
        return str(self.user) + '@' + self.service.url



from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse
from django.views.generic import DetailView, RedirectView, UpdateView
from django.contrib import messages
from django.utils.translation import ugettext_lazy as _

User = get_user_model()


class UserDetailView(LoginRequiredMixin, DetailView):

    model = User
    slug_field = "username"
    slug_url_kwarg = "username"


user_detail_view = UserDetailView.as_view()


class UserUpdateView(LoginRequiredMixin, UpdateView):

    model = User
    fields = ["name"]

    def get_success_url(self):
        return reverse("users:detail", kwargs={"username": self.request.user.username})

    def get_object(self):
        return User.objects.get(username=self.request.user.username)

    def form_valid(self, form):
        messages.add_message(
            self.request, messages.INFO, _("Infos successfully updated")
        )
        return super().form_valid(form)


user_update_view = UserUpdateView.as_view()


class UserRedirectView(LoginRequiredMixin, RedirectView):

    permanent = False

    def get_redirect_url(self):
        return reverse("users:detail", kwargs={"username": self.request.user.username})


user_redirect_view = UserRedirectView.as_view()



from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.chrome.options import Options

import time, sys

from django.test import LiveServerTestCase
from django.contrib.staticfiles.testing import StaticLiveServerTestCase

from django_sso_app.core.apps.services.models import Service


class TestFunzionale(StaticLiveServerTestCase):
    @classmethod
    def setUpClass(cls):
        for arg in sys.argv:
            if 'liveserver' in arg:
                cls.server_url = 'http://' + arg.split('=')[1]
                return
        super().setUpClass()
        cls.server_url = cls.live_server_url

    #def tearDownClass(cls):
    #    if cls.server_url == cls.live_server_url:
    #        super().tearDownClass()

    def setUp(self):
        #self.browser = webdriver.Chrome()
        
        options = webdriver.ChromeOptions()
        options.add_experimental_option('prefs', {'intl.accept_languages': 'en,en_US'})
        options.add_argument("start-maximized")
        options.add_argument("disable-infobars")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-dev-shm-usage")
        options.add_argument("--no-sandbox")
        options.add_experimental_option("useAutomationExtension", False)
        self.browser = webdriver.Chrome(options=options)
        
        #self.browser = webdriver.FirefoxDriver()
        self.browser.implicitly_wait(1)
        
        self.service = Service.objects.create(name="example", url="https://www.example.com")
        
    def tearDown(self):
        self.browser.close()

    def identifica_titolo(self, titolo):
        # riconosce la pagina del partito dal titolo
        self.assertIn(titolo, self.browser.title)

    def tag_contiene_testo(self, tag, testo, in_testa=False):
        """
        in_testa ==> testo == lista[id_lista][0].text
        """
        elementi = self.browser.find_elements_by_tag_name(tag)
        self.assertTrue(len(elementi)>0, 'lista vuota')
        if in_testa:
            self.assertEqual(
                testo, elementi[0].text,
                'Elemento "{}" non in testa'.format(testo)
            )
        else:
            self.assertIn(
                testo, [el.text for el in elementi]
            )

    def compila_input_by_name(self, name_input, testo, enter=True, el='input'):
        input_x = self.browser.find_element_by_xpath('//'+el+'[@name="'+name_input+'"]')

        input_x.send_keys(testo)
        if enter:
          input_x.send_keys(Keys.ENTER)

        return input_x

    def compila_input_by_id(self, eid, testo, enter=True):
        input_x = self.browser.find_element_by_id(eid)
        input_x.send_keys(testo)
        if enter:
          input_x.send_keys(Keys.ENTER)
        return input_x

    def click_link(self, url):
        link = self.browser.find_element_by_xpath('//a[@href="'+url+'"]')
        print('\n LINKS elementi', link)
        link.click()
        
    def find_by_tag_class_and_text(self, tag, css, text):
        return self.browser.find_element_by_xpath("//{0}[contains(@class, '{1}') and text()='{2}']".format(tag, css, text))

    def find_by_tag_and_name(self, tag, name):
        return self.browser.find_element_by_xpath('//{0}[@name="{1}"]'.format(tag, name))

    def find_by_tag_and_class(self, tag, cls):
        return self.browser.find_element_by_xpath("//{0}[contains(@class, '{1}')]".format(tag, cls))
      
    def press_next_button(self, i):
        self.browser.find_elements_by_xpath('//{0}[@name="{1}"]'.format('button', 'next_step'))[i].click()

if __name__ == '__main__':
    try:
        unittest.main(warnings='ignore')

    except KeyboardInterrupt:
        print('Ricevura terminazione da tastiera')
        exit(0)



from django.conf import settings

if settings.DJANGO_SSO_BACKEND_ENABLED:
    import json

    import jwt
    from django.urls import reverse
    from rest_framework import status

    from ...users.tests.test_views import UserTestCase


    class TestSubscriptions(UserTestCase):

        def test_user_can_list_service_subscriptions(self):
            """
            """
            response = self.client.post(
                reverse('rest_login'),
                data=json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.get(
                reverse('users:user_subscriptions', args=(self.user.sso_id, )),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_200_OK)
            self.assertEqual(len(response2.data['results']), 1)
            self.assertEqual(response2.data['results'][0].get('service').get('url'), self.service.url)


        def test_user_can_list_user_subscriptions(self):
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.get(
                reverse('users:user_subscriptions', args=(self.user.sso_id, )),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_200_OK)


        def test_user_can_retrieve_user_subscription(self):
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.get(
                reverse('users:user_subscription_detail', args=(self.user.sso_id, self.user_subscription.id,)),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_200_OK)
            self.assertEqual(response2.data.get('is_unsubscribed'), False)


        def test_user_can_unsubscribe_from_service_with_rev_incremented(self):
            first_rev = self.user.sso_rev
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.post(
                reverse('users:user_subscription_unsubscribe', args=(self.user.sso_id, self.user_subscription.id,)),
                data = json.dumps(self.valid_unsubscription),
                content_type='application/json'
            )
            cookies2 = response2.cookies

            print('\n\nRISPOSTA', response2.json())

            self.assertEqual(response2.status_code, status.HTTP_200_OK)

            self.assertEqual(response2.data.get('subscriptions')[0].get('is_unsubscribed'), True)

            self.user.refresh_from_db()
            second_rev = self.user.sso_rev

            self.assertEqual(second_rev, first_rev+1, 'rev not incremented on service unsubscription')

            jwt_cookie = cookies2.get('jwt')
            self.assertNotEqual(jwt_cookie, None, 'no new jwt received')

            unverified_payload = jwt.decode(jwt_cookie.value, None, False)

            print('\n\n\n Unv pay', unverified_payload)

            self.assertEqual(unverified_payload.get('rev'), second_rev, 'rev NOT updated in JWT')

        def test_user_can_unsubscribe_totally_with_rev_incremented(self):
            first_rev = self.user.sso_rev
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.post(
                reverse('users:user_unsubscription'),
                data = json.dumps(self.valid_unsubscription),
                content_type='application/json'
            )
            cookies2 = response2.cookies
            self.assertEqual(response2.status_code, status.HTTP_200_OK)

            print('\n\nRISPOSTA', response2.data)
            self.assertEqual(response2.data.get('is_unsubscribed'), True)

            self.user.refresh_from_db()
            second_rev = self.user.sso_rev

            self.assertEqual(second_rev, first_rev+1, 'rev not incremented on service unsubscription')

            jwt_cookie = cookies2.get('jwt')
            self.assertNotEqual(jwt_cookie, None, 'no cleared jwt received')


    class TestServices(UserTestCase):

        def test_user_can_list_services(self):
            """
            """
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.get(
                reverse('services:services_list'),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_200_OK)
            self.assertEqual(len(response2.data['results']), 2)

        def test_user_can_retrieve_service(self):
            """
            """
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.get(
                reverse('services:services_detail', args=(self.service.id, )),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_200_OK)
            self.assertEqual(response2.data.get('url'), self.service.url)


        def test_user_can_subscribe_to_service_with_rev_incremented_and_devices_deleted(self):
            """
            """
            first_rev = self.user.sso_rev
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies
            first_devices = self.user.devices.all()
            first_devices = [x.id for x in first_devices]

            response2 = self.client.post(
                reverse('services:services_subscription', args=(self.service2.id, )),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_201_CREATED)

            self.user.refresh_from_db()

            self.assertEqual(response2.data, 'Subscribed.')

            self.user.refresh_from_db()
            second_rev = self.user.sso_rev

            self.assertEqual(second_rev, first_rev+1, 'rev not incremented')

            response3 = self.client.get(
                reverse('services:services_list'),
                content_type='application/json'
            )

            self.assertEqual(response3.status_code, status.HTTP_200_OK)
            self.assertEqual(len(response3.data['results']), 2)

            cookies2 = response2.cookies
            jwt_cookie2 = cookies2.get('jwt').value
            unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

            print('\n\n\n Unv pay', unverified_payload2)

            self.assertEqual(unverified_payload2.get('rev'), second_rev, 'cookie rev NOT changed after login')

            second_devices = self.user.devices.all()
            second_devices = [x.id for x in second_devices]

            print('\n\n DEVICES', first_devices, second_devices)
            self.assertEqual(len(second_devices), 1)


        def test_user_can_not_subscribe_to_service_twice(self):
            """
            """
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.post(
                reverse('services:services_subscription', args=(self.service.id, )),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_409_CONFLICT)
            self.assertEqual(response2.data, 'Already subscribed.')


        def test_user_can_subscribe_to_service_from_alternative_url_with_rev_incremented(self):
            """
            """
            first_rev = self.user.sso_rev
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.post(
                reverse('users:user_subscriptions_create', args=(self.user.sso_id, self.service2.id, )),
                content_type='application/json'
            )
            print('\n\n\n RISPOSTA!', response2.data)
            self.assertEqual(response2.status_code, status.HTTP_201_CREATED)

            self.user.refresh_from_db()

            self.assertEqual(response2.data, 'Subscribed.')

            self.user.refresh_from_db()
            second_rev = self.user.sso_rev

            self.assertEqual(second_rev, first_rev+1, 'rev not incremented')

            response3 = self.client.get(
                reverse('services:services_list'),
                content_type='application/json'
            )

            self.assertEqual(response3.status_code, status.HTTP_200_OK)
            self.assertEqual(len(response3.data['results']), 2)

            cookies2 = response2.cookies
            jwt_cookie2 = cookies2.get('jwt').value
            unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

            print('\n\n\n Unv pay', unverified_payload2)

            self.assertEqual(unverified_payload2.get('rev'), second_rev, 'cookie rev NOT changed after login')



        def test_user_can_not_subscribe_to_service_from_alternative_url_if_not_same_user(self):
            """
            """
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_login),
                content_type='application/json'
            )
            cookies = response.cookies

            response2 = self.client.post(
                reverse('users:user_subscriptions_create', args=(self.user2.sso_id, self.service2.id, )),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_401_UNAUTHORIZED)


        def test_user_not_subscribe_to_service_from_alternative_url_if_not_same_user_but_staff_rev_is_incremented(self):
            """
            """
            first_rev = self.user.sso_rev

            response2 = self.client.post(
                reverse('users:user_subscriptions_create', args=(self.user.sso_id, self.service2.id, )),
                content_type='application/json',
                HTTP_AUTHORIZATION='Token {}'.format(self.staff_user_token.key)
            )
            print('\n\n\n RISPOSTA!', response2.data)

            self.assertEqual(response2.status_code, status.HTTP_201_CREATED)

            self.user.refresh_from_db()
            second_rev = self.user.sso_rev

            self.assertEqual(second_rev, first_rev+1, 'rev not incremented')

            cookies2 = response2.cookies
            jwt_cookie2 = cookies2.get('jwt')

            self.assertEqual(jwt_cookie2, None)




from .common import *


DJANGO_SSO_SERVER_DOMAIN = env('DJANGO_SSO_SERVER_DOMAIN', default='localhost')
DJANGO_SSO_SERVER_PORT = env.int('DJANGO_SSO_SERVER_PORT', default=8000)

ACCOUNT_DEFAULT_HTTP_PROTOCOL = 'http'
APP_DOMAIN = DJANGO_SSO_SERVER_DOMAIN + ':' + str(DJANGO_SSO_SERVER_PORT)

_SSO_DOMAINS = env('DJANGO_SSO_DOMAINS', default=DJANGO_SSO_SERVER_DOMAIN)
DJANGO_SSO_DOMAINS = _SSO_DOMAINS.split(',')
DJANGO_SSO_DOMAINS_DICT = {}
i=0
for el in DJANGO_SSO_DOMAINS:
    DJANGO_SSO_DOMAINS_DICT[el] = i
    i += 1

DJANGO_SSO_FULL_URLS_CHAIN = [ACCOUNT_DEFAULT_HTTP_PROTOCOL + '://' + el for el in DJANGO_SSO_DOMAINS]
DJANGO_SSO_URLS_CHAIN = [ACCOUNT_DEFAULT_HTTP_PROTOCOL + '://' + el for el in DJANGO_SSO_DOMAINS if el != APP_DOMAIN]
DJANGO_SSO_URLS_CHAIN_DICT = {}
i = 0
for el in DJANGO_SSO_URLS_CHAIN:
    DJANGO_SSO_URLS_CHAIN_DICT[el] = i
    i += 1

DJANGO_SSO_PASSEPARTOUT_PROCESS_ENABLED = len(DJANGO_SSO_URLS_CHAIN) > 0

DJANGO_SSO_DEFAULT_REFERRER = env('DJANGO_SSO_DEFAULT_REFERRER', default=ACCOUNT_DEFAULT_HTTP_PROTOCOL + '://' + APP_DOMAIN)

DJANGO_SSO_HIDE_PASSWORD_FROM_USER_SERIALIZER = env.bool('DJANGO_SSO_HIDE_PASSWORD_FROM_USER_SERIALIZER', default=True)

DJANGO_SSO_APIGW_ENABLED = env.bool('DJANGO_SSO_APIGW_ENABLED', default=False)


DJANGO_SSO_BACKEND_CORE_APPS = [
    # 'django_sso_app.core.apps.users',
    'django_sso_app.core.apps.groups',
    'django_sso_app.core.apps.services',
    'django_sso_app.core.apps.devices',
    'django_sso_app.core.apps.passepartout',

    DJANGO_SSO_APP_USER_PROFILE,
    DJANGO_SSO_FRONTEND_APP
]

if DJANGO_SSO_APIGW_ENABLED:
    DJANGO_SSO_BACKEND_CORE_APPS += [
        'django_sso_app.core.apps.api_gateway.kong',
    ]



import logging

import sentry_sdk

from sentry_sdk.integrations.django import DjangoIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.celery import CeleryIntegration


from .base import *  # noqa
from .base import env

# GENERAL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#secret-key
SECRET_KEY = env("DJANGO_SECRET_KEY")
# https://docs.djangoproject.com/en/dev/ref/settings/#allowed-hosts
ALLOWED_HOSTS = env.list("DJANGO_ALLOWED_HOSTS", default=["accounts.example.com"])

# DATABASES
# ------------------------------------------------------------------------------
DATABASES["default"] = env.db("DATABASE_URL")  # noqa F405
DATABASES["default"]["ATOMIC_REQUESTS"] = True  # noqa F405
DATABASES["default"]["CONN_MAX_AGE"] = env.int("CONN_MAX_AGE", default=60)  # noqa F405

# CACHES
# ------------------------------------------------------------------------------
CACHES = {
    "default": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": env("REDIS_URL"),
        "OPTIONS": {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
            # Mimicing memcache behavior.
            # http://niwinz.github.io/django-redis/latest/#_memcached_exceptions_behavior
            "IGNORE_EXCEPTIONS": True,
        },
    }
}

# SECURITY
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#secure-proxy-ssl-header
SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")
# https://docs.djangoproject.com/en/dev/ref/settings/#secure-ssl-redirect
SECURE_SSL_REDIRECT = env.bool("DJANGO_SECURE_SSL_REDIRECT", default=True)
# https://docs.djangoproject.com/en/dev/ref/settings/#session-cookie-secure
SESSION_COOKIE_SECURE = True
# https://docs.djangoproject.com/en/dev/ref/settings/#csrf-cookie-secure
CSRF_COOKIE_SECURE = True
# https://docs.djangoproject.com/en/dev/topics/security/#ssl-https
# https://docs.djangoproject.com/en/dev/ref/settings/#secure-hsts-seconds
# TODO: set this to 60 seconds first and then to 518400 once you prove the former works
SECURE_HSTS_SECONDS = 60
# https://docs.djangoproject.com/en/dev/ref/settings/#secure-hsts-include-subdomains
SECURE_HSTS_INCLUDE_SUBDOMAINS = env.bool(
    "DJANGO_SECURE_HSTS_INCLUDE_SUBDOMAINS", default=True
)
# https://docs.djangoproject.com/en/dev/ref/settings/#secure-hsts-preload
SECURE_HSTS_PRELOAD = env.bool("DJANGO_SECURE_HSTS_PRELOAD", default=True)
# https://docs.djangoproject.com/en/dev/ref/middleware/#x-content-type-options-nosniff
SECURE_CONTENT_TYPE_NOSNIFF = env.bool(
    "DJANGO_SECURE_CONTENT_TYPE_NOSNIFF", default=True
)

# MEDIA
# ------------------------------------------------------------------------------

# TEMPLATES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#templates
TEMPLATES[0]["OPTIONS"]["loaders"] = [  # noqa F405
    (
        "django.template.loaders.cached.Loader",
        [
            "django.template.loaders.filesystem.Loader",
            "django.template.loaders.app_directories.Loader",
        ],
    )
]

# EMAIL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#default-from-email
DEFAULT_FROM_EMAIL = env(
    "DJANGO_DEFAULT_FROM_EMAIL", default="Django SSO App <noreply@accounts.example.com>"
)
# https://docs.djangoproject.com/en/dev/ref/settings/#server-email
SERVER_EMAIL = env("DJANGO_SERVER_EMAIL", default=DEFAULT_FROM_EMAIL)
# https://docs.djangoproject.com/en/dev/ref/settings/#email-subject-prefix
EMAIL_SUBJECT_PREFIX = env(
    "DJANGO_EMAIL_SUBJECT_PREFIX", default="[Django SSO App]"
)

# ADMIN
# ------------------------------------------------------------------------------
# Django Admin URL regex.
ADMIN_URL = env("DJANGO_ADMIN_URL")

# Anymail (Mailgun)
# ------------------------------------------------------------------------------
# https://anymail.readthedocs.io/en/stable/installation/#installing-anymail
INSTALLED_APPS += ["anymail"]  # noqa F405
EMAIL_BACKEND = "anymail.backends.mailgun.EmailBackend"
# https://anymail.readthedocs.io/en/stable/installation/#anymail-settings-reference
ANYMAIL = {
    "MAILGUN_API_KEY": env("MAILGUN_API_KEY"),
    "MAILGUN_SENDER_DOMAIN": env("MAILGUN_DOMAIN"),
    "MAILGUN_API_URL": env("MAILGUN_API_URL", default="https://api.mailgun.net/v3"),
}

# Collectfast
# ------------------------------------------------------------------------------
# https://github.com/antonagestam/collectfast#installation
INSTALLED_APPS = ["collectfast"] + INSTALLED_APPS  # noqa F405
AWS_PRELOAD_METADATA = True

# LOGGING
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#logging
# See https://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.

LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "verbose": {
            "format": "%(levelname)s %(asctime)s %(module)s "
            "%(process)d %(thread)d %(message)s"
        }
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        }
    },
    "root": {"level": "INFO", "handlers": ["console"]},
    "loggers": {
        "django.db.backends": {
            "level": "ERROR",
            "handlers": ["console"],
            "propagate": False,
        },
        # Errors logged by the SDK itself
        "sentry_sdk": {"level": "ERROR", "handlers": ["console"], "propagate": False},
        "django.security.DisallowedHost": {
            "level": "ERROR",
            "handlers": ["console"],
            "propagate": False,
        },
    },
}

# Sentry
# ------------------------------------------------------------------------------
SENTRY_DSN = env("SENTRY_DSN")
SENTRY_LOG_LEVEL = env.int("DJANGO_SENTRY_LOG_LEVEL", logging.INFO)

sentry_logging = LoggingIntegration(
    level=SENTRY_LOG_LEVEL,  # Capture info and above as breadcrumbs
    event_level=logging.ERROR,  # Send errors as events
)
sentry_sdk.init(
    dsn=SENTRY_DSN,
    integrations=[sentry_logging, DjangoIntegration(), CeleryIntegration()],
)

# Your stuff...
# ------------------------------------------------------------------------------



from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from ..services.views import SubscriptionApiViewSet
from .views import (CheckUserExistence,
                    UserApiViewSet,
                    UserUnsubscriptionApiView,
                    UserRevisionsApiView)

_urlpatterns = [
    url(r'^$', UserApiViewSet.as_view({'get': 'list'}), name="list"),

    url(r'^(?P<pk>[0-9A-Fa-f-]+)/$',
        UserApiViewSet.as_view({'get': 'retrieve', 'patch': 'partial_update'}, name="detail"),
        name='detail'),


    url(r'^check/$',
        CheckUserExistence.as_view(), name="check_user_existence"),

    url(r'^revisions/$',
        UserRevisionsApiView.as_view(), name="users_revision"),

    url(r'^unsubscribe/$',
        UserUnsubscriptionApiView.as_view(),
        name='user_unsubscription'),


    url(r'^(?P<user_pk>[0-9A-Fa-f-]+)/subscriptions/$',
        SubscriptionApiViewSet.as_view({'get': 'list'}),
        name='user_subscriptions'),

    url(r'^(?P<user_pk>[0-9A-Fa-f-]+)/subscriptions/create/(?P<service_pk>\d+)/$',
        SubscriptionApiViewSet.as_view({'post': 'create'}),
        name='user_subscriptions_create'),

    url(r'^(?P<user_pk>[0-9A-Fa-f-]+)/subscriptions/(?P<pk>\d+)/$',
        SubscriptionApiViewSet.as_view({'get': 'retrieve'}),
        name='user_subscription_detail'),

    url(r'^(?P<user_pk>[0-9A-Fa-f-]+)/subscriptions/(?P<pk>\d+)/unsubscribe/$',
        SubscriptionApiViewSet.as_view({'post': 'unsubscribe'}),
        name='user_subscription_unsubscribe'),

]

urlpatterns = format_suffix_patterns(_urlpatterns)



import logging

from django.apps import AppConfig
from django.conf import settings

from django_sso_app.core.utils import get_profile_model

logger = logging.getLogger("profiles")


class ProfilesConfig(AppConfig):
    name = 'django_sso_app.core.apps.profiles'

    def ready(self):
        Profile = get_profile_model()

        for el in settings.DJANGO_SSO_REQUIRED_PROFILE_FIELDS:
            if getattr(Profile, el, None) is None:
                raise Exception(
                    'App profile {0} is missing {1} field'.format(Profile,
                                                                  el))

        try:
            import django_sso_app.core.apps.profiles.signals  # noqa F401
        except ImportError:
            pass

        super(ProfilesConfig, self).ready()
        logger.info("django-sso-app profiles app ready!")



from django.conf import settings

import json

import jwt
from allauth.account.models import EmailAddress
from allauth.account.adapter import get_adapter
from django.contrib.auth.hashers import make_password
from django.contrib.auth import get_user_model
from django.test import TestCase, Client
from django.urls import reverse
from rest_framework import status
from rest_framework.authtoken.models import Token

from ...services.models import Service, Subscription
from django_sso_app.core.utils import get_profile_model

User = get_user_model()
Profile = get_profile_model()


class ProfileTestCase(TestCase):
    def setUp(self):
        self.client = Client()

        email = self.email = "pippo@disney.com"
        username = self.username = "pippo"
        password = self.password = "paperina"
        updated_password = "pippapippa"


        username2 = "cenerentola"
        email2 = "cenerentola@disney.com"
        password2 = "tanticassi"

        self.user = User.objects.create_user(username=username, email=email, password=password)
        self.user_email = EmailAddress.objects.create(user=self.user,
                                                      email=email,
                                                      primary=True,
                                                      verified=True)

        self.user2 = User.objects.create_user(username=username2, email=email2, password=password2)
        self.user2_email = EmailAddress.objects.create(user=self.user2,
                                                      email=email2,
                                                      primary=True,
                                                      verified=True)

        self.service = Service.objects.create(url='http://example.com')
        self.user_subscription = Subscription.objects.create(user=self.user,
                                                             service=self.service)
        self.service2 = Service.objects.create(url='http://disney.org')

        self.staff_user_username = 'staff'
        self.staff_user_password = 'abc123456'
        self.staff_user = User.objects.create_user(username=self.staff_user_username, email='staff@example.com', password=self.staff_user_password, is_staff=True)
        self.staff_user_email = EmailAddress.objects.create(user=self.staff_user,
                                                      email=self.staff_user.email,
                                                      primary=True,
                                                      verified=True)
        self.staff_user_token = Token.objects.create(user=self.staff_user)

        self.valid_new_profile = {
            'email': 'paperino@gmail.com',
            'username': 'paperino',
            'password': 'nonnapapera',
            'first_name': 'Paolino',
            'referrer': 'http://example.com'
        }

        self.valid_new_profile_hashed_pw = {
            'email': 'paperino@gmail.com',
            'username': 'paperino',
            'password': make_password('nonnapapera'),
            'first_name': 'Paolino',
            'referrer': 'http://example.com'
        }

        self.valid_new_profile_login = {
            'username': 'paperino',
            'password': 'nonnapapera',
            'fingerprint': '123456'
        }


        self.staff_user_valid_login = {
            'username': self.staff_user.username,
            'password': self.staff_user_password,
            'fingerprint': '123456'
        }

        self.valid_login = {
            'username': username,
            'password': password,
            'fingerprint': '123456'
        }

        self.valid_staff_login = {
            'username': 'staff',
            'password': self.staff_user_password,
            'fingerprint': '123456'
        }

        self.valid_unsubscription = {
            'password': password
        }

        self.valid_login_after_password_change = {
            'username': username,
            'password': updated_password,
            'fingerprint': '123456'
        }

        self.valid_profile_update = {
            'first_name': 'Pippo'
        }

        self.valid_password_update = {
            'password': updated_password
        }

        self.valid_email_update = {
            'email': 'pippo2@gmail.com'
        }

        adapter = get_adapter()
        self.device0 = adapter.add_user_device(self.user, '000000')
        self.device1 = adapter.add_user_device(self.user, '111111')

        self.valid_registration = {
            'email': 'topolino@disney.com',
            'username': 'topolino',
            'password1': 'paperina',
            'password2': 'paperina',
            'first_name': 'Mickey',
            'last_name': 'Mouse',
            'country': 'Italia',
            'latitude': 12.345,
            'longitude': 54.321,
            'referrer': 'http://example.com'
        }


class TestProfileRetrieve(ProfileTestCase):

    def test_user_can_retrieve_profile_with_subscriptions(self):
        """
          Registration via rest_auth view
        """
        print('FIRST', json.dumps(self.valid_login))

        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies
        print('RISPSTA1', response.data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = self.client.get(
            reverse('users:detail', args=(self.user.sso_id, )),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        print('RISPSTA2', response2.data)

        self.assertEqual(response2.data.get('subscriptions'), [{'url': el.service.url, 'is_unsubscribed': el.is_unsubscribed} for el in self.user.subscriptions.all()])



class TestProfileUpdate(ProfileTestCase):

    def test_profile_update_refreshes_and_returns_current_device_jwt_and_deletes_other_devices(self):
        """
        Profile update refreshes caller jwt and deletes all but caller devices
        """
        first_rev = self.user.sso_rev
        response = self.client.post(
            reverse('rest_login'),
            data = json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies
        print('\n\n\nRESPONSE1!!!!!!\n\n', response.json())

        response2 = self.client.patch(
            '/api/v1/profiles/{0}/'.format(self.user.sso_id),
            data = json.dumps(self.valid_profile_update),
            content_type='application/json'
        )
        cookies2 = response2.cookies

        print('\n\n\nRESPONSE!!!!!!\n\n', response2.json())

        self.user.refresh_from_db()
        second_rev = self.user.sso_rev
        self.assertEqual(second_rev, first_rev+1, 'rev not incremented')

        self.assertTrue(len(cookies2.keys()) > 0, 'no jwt cookie set')

        devices = self.user.devices.all()

        self.assertEqual(devices.count(), 1, 'other devices not deleted')

        self.assertEqual(devices.first().fingerprint, self.valid_login['fingerprint'], 'device not updated')

        jwt_cookie = cookies.get('jwt').value
        jwt_cookie2 = cookies2.get('jwt').value

        # print('\nRISPOSTA, user rev', self.user.sso_rev, '\n\nTOKEN', jwt_cookie)

        unverified_payload = jwt.decode(jwt_cookie, None, False)
        unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

        self.assertEqual(unverified_payload2.get('rev'), unverified_payload.get('rev')+1, 'rev not incremented in cookie')


    def test_profile_update_by_staff_deletes_all_profile_devices(self):
        """
        profile update by staff refreshes deletes all user devices
        """

        first_rev = self.user.sso_rev
        response2 = self.client.patch(
            '/api/v1/profiles/{0}/'.format(self.user.sso_id),
            data = json.dumps(self.valid_profile_update),
            content_type='application/json',
            HTTP_AUTHORIZATION='Token {}'.format(self.staff_user_token.key)
        )
        cookies2 = response2.cookies

        self.assertTrue(len(cookies2.keys()) == 0, 'jwt cookie set')

        self.user.refresh_from_db()
        second_rev = self.user.sso_rev

        self.assertEqual(second_rev, first_rev+1, 'rev not incremented')

        devices = self.user.devices.all()

        self.assertEqual(devices.count(), 0, 'user devices not deleted')



import logging

from django.conf import settings
from django.contrib.auth.models import Group as GroupModel
from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from .models import Group
from ..api_gateway.functions import create_consumer_acl, delete_consumer_acl

logger = logging.getLogger('groups')


def create_user_group(user, group):
    logger.info('creating user group {} for {}'.format(group, user))
    if not settings.TESTING_MODE:
        status_code, resp = create_consumer_acl(user, group.name)
        print('CREATED', status_code, resp)


def delete_user_group(user, group):
    logger.info('deleting user group {} for {}'.format(group, user))
    if not settings.TESTING_MODE:
        status_code, resp = delete_consumer_acl(user, group.name)
        print('DELETED', status_code)


@receiver(m2m_changed)
def signal_handler_when_user_is_added_or_removed_from_group(action, instance, pk_set, model, **kwargs):
    if model == GroupModel:
        user = instance
        user_updated = False
        if action == 'pre_add':
            user_updated = True
            for pk in pk_set:
                group = Group.objects.get(id=pk)
                create_user_group(user, group)
        elif action == 'pre_remove':
            user_updated = True
            for pk in pk_set:
                group = Group.objects.get(id=pk)
                delete_user_group(user, group)
        if user_updated:
            user.update_rev(True)



import json
import logging

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.db import transaction
from django.utils.encoding import smart_str

from django_sso_app.core.utils import get_profile_model

logger = logging.getLogger("app")


def get_sso_id(consumer_username):
    return consumer_username.replace('_', '-')


def get_user_groups_from_header(groups_header):
    if groups_header is not None and groups_header != '':
        return list(map(str.strip, groups_header.split(',')))
    return []


def update_user_groups(user, groups):
    for group in groups:
        g, _created = Group.objects.get_or_create(name=group)
        if _created:
            logger.info('Created group {}'.format(g))
        g.user_set.add(user)


@transaction.atomic
def create_local_profile_from_headers(request):
    consumer_username = request.META.get(
        settings.DJANGO_SSO_CONSUMER_USERNAME_HEADER)
    consumer_groups = request.META.get(
        settings.DJANGO_SSO_CONSUMER_GROUPS_HEADER, None)

    logger.info("Creating local profile for user {} with groups {} ..."
                .format(smart_str(consumer_username),
                        consumer_groups))

    sso_id = get_sso_id(consumer_username)

    # creating new user
    new_user = get_user_model().objects.create(
        username=sso_id
    )
    if settings.DJANGO_SSO_APP_USES_EXTERNAL_USER_PROFILE:
        new_user_profile = get_profile_model().objects.get_or_create(
            user=new_user, sso_id=sso_id, sso_rev=0)
    else:
        new_user.sso_id = sso_id
        new_user.save()

    if settings.DJANGO_SSO_MANAGE_USER_GROUPS:
        new_user_groups = get_user_groups_from_header(consumer_groups)

        is_staff = ('staff' in new_user_groups)
        if (is_staff):
            logger.info('New user "{}" is "staff"')

        update_user_groups(new_user, new_user_groups)

    logger.info("Created local profile for user {} with SSO ID {}"
                .format(smart_str(consumer_username),
                        sso_id))

    return new_user


class DjangoSsoAppUserStatus(object):
    def __init__(self, is_active, is_unsubscribed, is_to_subscribe):
        self.is_active = is_active
        self.is_unsubscribed = is_unsubscribed
        self.is_to_subscribe = is_to_subscribe

    def __repr__(self):
        return ("{}(is_active={}, is_unsubscribed={}, is_to_subscribe={})"
                .format(self.__class__.__name__, self.is_active,
                        self.is_unsubscribed, self.is_to_subscribe))

    @staticmethod
    def get_user_status(is_unsubscribed, subscriptions, email_verified=True):
        _is_active = settings.DJANGO_SSO_DEACTIVATE_USER_ON_UNSUBSCRIPTION

        if is_unsubscribed:  # is unsubscribed from sso
            return DjangoSsoAppUserStatus(is_active=_is_active,
                                          is_unsubscribed=True,
                                          is_to_subscribe=False)
        else:  # is still subscribed to sso
            for subscription in subscriptions:
                if subscription[
                    "url"] == settings.DJANGO_SSO_SERVICE_URL:  #
                    # subscription found
                    if subscription[
                        "is_unsubscribed"]:  # user is unsubscribed from service
                        return DjangoSsoAppUserStatus(is_active=(_is_active
                                                                 and email_verified),
                                                      is_unsubscribed=True,
                                                      is_to_subscribe=False)
                    return DjangoSsoAppUserStatus(is_active=email_verified,
                                                  is_unsubscribed=False,
                                                  is_to_subscribe=False)

            # is NOT subscribed
            return DjangoSsoAppUserStatus(
                is_active=email_verified, is_unsubscribed=False,
                is_to_subscribe=settings
                    .DJANGO_SSO_SERVICE_SUBSCRIPTION_IS_MANDATORY)


def check_profile_on_sso_by_username(username):
    logger.info("Checking with SSO Backend if user {} already exists ..."
                .format(smart_str(username)))

    url = settings.DJANGO_SSO_BACKEND_USERS_CHECK_URL
    params = {
        "username": username
    }
    response = requests.get(url=url, params=params, timeout=10, verify=False)

    if response.status_code == requests.codes.NOT_FOUND:
        logger.info("User {} does NOT exist".format(smart_str(username)))

        return None

    response.raise_for_status()
    sso_id = json.loads(response.text).get("id", None)

    logger.info("User {} already exists with SSO ID {}"
                .format(smart_str(username), sso_id))

    return sso_id


def check_profile_on_sso_by_email(email):
    logger.info("Checking with SSO if a user with email {} already exists ..."
                .format(smart_str(email)))

    url = settings.DJANGO_SSO_BACKEND_USERS_CHECK_URL
    params = {
        "email": email
    }
    response = requests.get(url=url, params=params, timeout=10, verify=False)

    if response.status_code == requests.codes.NOT_FOUND:
        logger.info("A user with email {} does NOT exist"
                    .format(smart_str(email)))

        return None

    response.raise_for_status()
    sso_id = json.loads(response.text).get("id", None)

    logger.info("A user with email {} already exists with SSO ID {}"
                .format(smart_str(email), sso_id))

    return sso_id


def get_profile_from_sso(sso_id, encoded_jwt):
    logger.info("Getting SSO profile for ID {} ...".format(sso_id))

    url = settings.DJANGO_SSO_BACKEND_USER_URL.format(
        user_id=sso_id) + '?with_password=true'
    headers = {
        "Authorization": "Bearer {jwt}".format(jwt=encoded_jwt)
    }
    response = requests.get(url=url, headers=headers, timeout=10, verify=False)
    response.raise_for_status()
    sso_profile = response.json()

    logger.info("Retrieved SSO profile for ID {}".format(sso_id))

    return sso_profile


@transaction.atomic
def create_local_profile_from_sso(sso_profile, can_subscribe=False):
    logger.info("Creating local profile for user {} with SSO ID {} ..."
                .format(smart_str(sso_profile["username"]), sso_profile["id"]))

    profile = _create_or_update_local_profile_from_sso(
        sso_profile=sso_profile, can_subscribe=can_subscribe)

    logger.info("Created local profile for user {} with SSO ID {}"
                .format(smart_str(sso_profile["username"]), sso_profile["id"]))

    return profile


@transaction.atomic
def update_local_profile_from_sso(sso_profile, profile, can_subscribe=False):
    assert profile is not None

    logger.info("Updating local profile for user {} with SSO ID {} ..."
                .format(smart_str(sso_profile["username"]), sso_profile["id"]))

    _create_or_update_local_profile_from_sso(sso_profile=sso_profile,
                                             profile=profile,
                                             can_subscribe=can_subscribe)

    logger.info("Updated local profile for user {} with SSO ID {}"
                .format(smart_str(sso_profile["username"]), sso_profile["id"]))


@transaction.atomic
def _create_or_update_local_profile_from_sso(sso_profile, profile=None,
                                             can_subscribe=False):
    username = sso_profile["username"]
    email = sso_profile["email"]
    email_verified = sso_profile["email_verified"]
    password = sso_profile["password"]
    date_joined = sso_profile["date_joined"]
    first_name = sso_profile["first_name"]
    last_name = sso_profile["last_name"]
    sso_id = sso_profile["id"]

    creating = profile is None

    if creating:
        sso_user_status = DjangoSsoAppUserStatus.get_user_status(
            is_unsubscribed=sso_profile["is_unsubscribed"],
            subscriptions=sso_profile["subscriptions"])

        logger.info("Status for user {} with SSO ID {}: {}"
                    .format(smart_str(username), sso_id, sso_user_status))

        user_model = get_user_model()
        profile_model = get_profile_model()
        user = user_model()
        user.username = username
        profile = profile_model()
        profile.sso_id = sso_id
    else:
        sso_user_status = DjangoSsoAppUserStatus.get_user_status(
            is_unsubscribed=sso_profile["is_unsubscribed"],
            subscriptions=sso_profile["subscriptions"],
            email_verified=email_verified)

        logger.info("Status for user {} with SSO ID {}: {}"
                    .format(smart_str(username), sso_id, sso_user_status))

        user = profile.user

    is_active = sso_user_status.is_active
    is_unsubscribed = sso_user_status.is_unsubscribed
    is_to_subscribe = sso_user_status.is_to_subscribe  # ! and not
    # is_unsubscribed

    sso_profile['sso_id'] = sso_id
    sso_profile['sso_rev'] = sso_profile['rev']

    user.email = email
    user.password = password
    user.date_joined = date_joined
    user.first_name = first_name
    user.last_name = last_name
    user.is_active = is_active

    for f in settings.DJANGO_SSO_REQUIRED_PROFILE_FIELDS:
        setattr(profile, f, sso_profile[f])

    user.save()
    if creating:
        profile.user = user
    profile.save()

    setattr(user, "is_to_subscribe", is_to_subscribe)
    if is_to_subscribe and can_subscribe:
        _subscribe_to_service(sso_id)

    return profile


def _subscribe_to_service(sso_id):
    logger.info("Subscribing user with SSO ID {sso_id} to service"
                " {service_name} ..."
                .format(sso_id=sso_id,
                        service_name=settings.DJANGO_SSO_SERVICE_URL))

    headers = {
        "Content-Type": "application/json",
        "Authorization": "Bearer {jwt}".format(
            jwt=settings.DJANGO_SSO_STAFF_JWT),
    }

    # Get all available services
    url = settings.DJANGO_SSO_BACKEND_SERVICES_URL
    response = requests.get(url=url, headers=headers, timeout=10, verify=False)
    response.raise_for_status()
    sso_services = response.json()

    # Find out the id for current service
    service_id = None
    for sso_service in sso_services:
        if sso_service["url"] == settings.DJANGO_SSO_SERVICE_URL:
            service_id = sso_service["id"]
            break
    if not service_id:
        _msg = ("Current service {} is not listed in SSO!"
                .format(settings.DJANGO_SSO_SERVICE_URL))
        raise Exception(_msg)

    # Subscribe to current service
    url = (settings.DJANGO_SSO_BACKEND_USER_SUBSCRIPTIONS_CREATE_URL
           .format(user_id=sso_id, service_id=service_id))
    response = requests.post(url=url, headers=headers, timeout=10, verify=False)
    response.raise_for_status()
    logger.info("User with SSO ID {sso_id} was successfully subscribed to"
                " service {service_name}!"
                .format(sso_id=sso_id,
                        service_name=settings.DJANGO_SSO_SERVICE_URL))



import logging

from rest_framework import viewsets

from django_sso_app.backend.utils import get_request_jwt, update_response_jwt
from django_sso_app.core.permissions import is_staff
from django_sso_app.core.utils import get_profile_model
from .serializers import ProfileSerializer, ProfilePublicSerializer

logger = logging.getLogger('profiles')
Profile = get_profile_model()


class ProfileViewSet(viewsets.ModelViewSet):
    lookup_field = 'sso_id'

    def get_queryset(self):
        user = self.request.user
        if is_staff(user):
            return Profile.objects.all()
        else:
            return Profile.objects.filter(user=user)

    def get_serializer_class(self):
        if self.request.user.is_authenticated:
            return ProfileSerializer
        else:
            return ProfilePublicSerializer

    def list(self, request, *args, **kwargs):
        """
        List profiles
        """
        return super(ProfileViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        Profile detail
        """
        return super(ProfileViewSet, self).retrieve(request, *args, **kwargs)

    def partial_update(self, request, *args, **kwargs):
        """
        Profile update
        """
        response = super(ProfileViewSet, self).partial_update(request, *args, **kwargs)
        response_status_code = response.status_code
        logger.info('Profile partial_update status code is {}, {}'.format(response_status_code, request.data))

        if response_status_code == 200:
            received_jwt = get_request_jwt(request)
            logger.info('Updating JWT: {}'.format(received_jwt))

            profile = self.get_object()
            user = profile.user
            requesting_user = self.request.user

            from_browser_as_same_user = received_jwt is not None and requesting_user == user
            #logger.info('From browser as same user? {} == {}? {}'. format(requesting_user, user, requesting_user.id == user.id))

            if from_browser_as_same_user: # From browser as same user
                logger.info('User {0} updated profile, updating response JWT'.format(user))
                update_response_jwt(received_jwt, user, request, response)

        return response



import logging
from urllib.parse import urlencode

from allauth.account.adapter import get_adapter
from django.conf import settings
from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from rest_auth.registration.views import RegisterView as rest_authRegisterView
from rest_auth.registration.views import \
    VerifyEmailView as rest_authVerifyEmailView
from rest_auth.views import LoginView as rest_authLoginView
from rest_auth.views import LogoutView as rest_authLogoutView
from rest_auth.views import PasswordChangeView as rest_authPasswordChangeView
from rest_auth.views import \
    PasswordResetConfirmView as rest_authPasswordResetConfirmView
from rest_auth.views import PasswordResetView as rest_authPasswordResetView
from rest_framework import status
from rest_framework.generics import RetrieveAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.reverse import reverse

from django_sso_app.core.apps.passepartout.models import Passepartout
from django_sso_app.core.apps.users.serializers import UserSerializer
from django_sso_app.core.permissions import is_staff
from ..handlers import jwt_encode
from ..serializers import JWTSerializer
from ..utils import set_cookie, invalidate_cookie, get_request_jwt_fingerprint

logger = logging.getLogger('backend')


# rest_auth views

class RegisterAPIView(rest_authRegisterView):
    """
    Register a new user
    """

    def create(self, request, *args, **kwargs):
        try:
            logger.info('Registering user {}'.format(request.data))
            return super(RegisterAPIView, self).create(request, *args, **kwargs)

        except:
            logger.exception('Registration error')
            raise


class VerifyEmailAPIView(rest_authVerifyEmailView):
    """
    Verify new user email
    """

    def post(self, request, *args, **kwargs):
        logger.info('Verifying email')
        try:
            return super(VerifyEmailAPIView, self).post(request, *args,
                                                        **kwargs)
        except:
            logger.exception('Email verification error')
            raise


class LoginAPIView(rest_authLoginView):
    """
    User login
    """

    def return_unauthorized_if_user_is_staff(self):
        if self.user.is_staff:
            logger.warning('Staff user {0} tried to login'.format(self.user))
            return Response('Login unauthorized',
                            status=status.HTTP_403_FORBIDDEN)
        return None

    def get_device(self, fingerprint):
        device = self.user.devices.filter(fingerprint=fingerprint).first()
        adapter = get_adapter(self.request)

        if device is None:
            logger.info('User {0} Login with new Device fingerprint {1}'.format(
                self.user, fingerprint))
            device = adapter.add_user_device(self.user, fingerprint)

        assert device is not None

        setattr(self, 'device', device)

        return device

    def get_response(self):
        # check uesr is not staff
        failing_response = self.return_unauthorized_if_user_is_staff()
        if failing_response is not None:
            return failing_response

        passepartout_url = None

        if settings.DJANGO_SSO_PASSEPARTOUT_PROCESS_ENABLED:
            logger.info(
                'Passepartout login enabled, DJANGO_SSO_URLS_CHAIN: {0}'.format(
                    settings.DJANGO_SSO_URLS_CHAIN))
            nextUrl = self.request.GET.get('next', None)
            if nextUrl is None:
                nextUrl = settings.APP_URL

            fingerprint = self.serializer.validated_data.get('fingerprint', self.request.data.get('fingerprint',
                                                                                                  'undefined'))
            device = self.get_device(fingerprint)

            logger.info(
                'Login started passepartout logic'
                'for user {0}, with device {1}, nextUrl is {2}'.format(
                    self.user, device, nextUrl))

            passepartout = Passepartout.objects.create_passepartout(
                device=device, jwt=self.token)
            logger.info('Created passepartout object {0}'.format(passepartout))

            args = {
                'next': nextUrl
            }
            next_sso_service = settings.DJANGO_SSO_URLS_CHAIN[0]

            redirect_to = next_sso_service + reverse(
                'passepartout:passepartout_login',
                args=(passepartout.token,)) + '?' + urlencode(args)

            passepartout_url = redirect_to

        else:
            logger.info('Only one SSO instance, no passepartout process')

        data = {
            'user': self.user,
            'token': self.token,
            'passepartout_redirect_url': passepartout_url
        }
        serializer = JWTSerializer(instance=data,
                                   context={'request': self.request})

        response = Response(serializer.data, status=status.HTTP_200_OK)
        set_cookie(response, 'jwt', self.token, None)

        return response

    def login(self, *args, **kwargs):
        self.user = self.serializer.validated_data['user']

        if is_staff(self.user):
            logger.info(
                'User {0} is staff, returning.'.format(self.user))
            return

        # print('\n\nCI SONO!!', self.request.data)
        #print('DATAAAS??', self.request.data, self.request.data.get('fingerprint'))
        # fingerprint = self.serializer.validated_data['fingerprint']
        fingerprint = self.serializer.validated_data.get('fingerprint', self.request.data.get('fingerprint'))

        logger.info(
            'Start logging in user {0} with fingerprint {1}'.format(self.user,
                                                                    fingerprint))

        if self.user.apigw_consumer_id is None:
            logger.info(
                'User {} is logging in for the first time, creating apigw '
                'consumer'.format(
                    self.user))
            adapter = get_adapter()
            adapter.create_apigw_consumer(self.user)

        device = self.get_device(fingerprint)

        self.token = jwt_encode(self.user, device=device)

        logger.info('User {0} is logging in with device {1}'.format(self.user,
                                                                    self.device))

        if getattr(settings, 'REST_SESSION_LOGIN', True):
            self.process_login()

    def post(self, request, *args, **kwargs):
        try:
            return super(LoginAPIView, self).post(request, *args, **kwargs)
        except:
            logger.exception('Login error')
            raise


class LogoutAPIView(rest_authLogoutView):
    """
    Calls Django logout method and deletes the Token object
    assigned to the current User object.

    post:
    Accepts "next" url parameter
    Returns a JSON object containing logout detail message and, if there are
    more than one django-sso instances, a passepartout redirect url.
    """

    def logout(self, request):

        if settings.DJANGO_SSO_PASSEPARTOUT_PROCESS_ENABLED:
            logger.info(
                'Passepartout logout enabled, DJANGO_SSO_URLS_CHAIN: {0}'.format(
                    settings.DJANGO_SSO_URLS_CHAIN))
            nextUrl = self.request.GET.get('next', None)
            if nextUrl is None:
                nextUrl = settings.APP_URL

            args = {
                'next': nextUrl
            }

            requesting_device_fingerprint = get_request_jwt_fingerprint(request)
            device = request.user.devices.filter(
                fingerprint=requesting_device_fingerprint).first()

            assert device is not None

            next_sso_service = settings.DJANGO_SSO_URLS_CHAIN[0]
            redirect_to = next_sso_service + reverse(
                'passepartout:passepartout_logout',
                args=[device.id]) + '?' + urlencode(args)

            response_object = {"detail": _("Successfully logged out."),
                               "passepartout_redirect_url": redirect_to}
            response = Response(response_object, status=status.HTTP_200_OK)
        else:
            if not settings.DEBUG:
                adapter = get_adapter(self.request)

                try:
                    adapter.delete_request_device(request)
                except KeyError as e:
                    return Response(str(e), status=status.HTTP_400_BAD_REQUEST)

            response = super(LogoutAPIView, self).logout(request)

        invalidate_cookie(response)

        return response

    def post(self, request, *args, **kwargs):
        try:
            return super(LogoutAPIView, self).post(request, *args, **kwargs)
        except:
            logger.exception('Logout error')
            raise


class UserDetailsAPIView(RetrieveAPIView):
    """
    Overrides django-rest-auth default UserDetailsView
    in order to prevent PATCH
    """

    serializer_class = UserSerializer
    permission_classes = (IsAuthenticated,)

    def get_object(self):
        return self.request.user

    def get_queryset(self):
        """
        Adding this method since it is sometimes called when using
        django-rest-swagger
        https://github.com/Tivix/django-rest-auth/issues/275
        """
        return get_user_model().objects.none()

    def get(self, request, *args, **kwargs):
        try:
            return super(UserDetailsAPIView, self).get(request, *args, **kwargs)
        except:
            logger.exception('Error getting user details')
            raise


class PasswordResetAPIView(rest_authPasswordResetView):
    """
    Reset password
    """

    def post(self, request, *args, **kwargs):
        logger.info('Asking for password reset')
        try:
            return super(PasswordResetAPIView, self).post(request, *args,
                                                          **kwargs)
        except:
            logger.exception('Error while asking password reset')
            raise


class PasswordResetConfirmAPIView(rest_authPasswordResetConfirmView):
    """
    Password reset confirmation
    """

    def post(self, request, *args, **kwargs):
        logger.info('Confirming password reset')
        try:
            return super(PasswordResetConfirmAPIView, self).post(request, *args,
                                                                 **kwargs)
        except:
            logger.exception('Error resetting password')
            raise


class PasswordChangeAPIView(rest_authPasswordChangeView):
    """
    Password change
    """

    def post(self, request, *args, **kwargs):
        logger.info('Changing password')
        try:
            return super(PasswordChangeAPIView, self).post(request, *args,
                                                           **kwargs)
        except:
            logger.exception('Error changing password')
            raise



import uuid

from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _


class UuidPKModel():
    class Meta:
        abstract = True

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)


class DeactivableModel(models.Model):
    class Meta:
        abstract = True

    is_active = models.BooleanField(_("is active"), default=True)


class CreatedAtModel(models.Model):
    class Meta:
        abstract = True

    created_at = models.DateTimeField(_("created at"),
                                      editable=False,
                                      auto_now_add=True)


class UpdatableModel(models.Model):
    class Meta:
        abstract = True

    updated_at = models.DateTimeField(_("updated at"),
                                      editable=False,
                                      null=True,
                                      blank=True)

    def save(self, *args, **kwargs):
        if not self._state.adding:
            self.updated_at = timezone.now()
        return super(UpdatableModel, self).save(*args, **kwargs)


class PublicableModel(models.Model):
    class Meta:
        abstract = True

    is_public = models.BooleanField(default=False)


class TimespanModel(models.Model):
    class Meta:
        abstract = True

    started_at = models.DateTimeField(_("started at"), null=True, blank=True)

    ended_at = models.DateTimeField(_("ended at"), null=True, blank=True)



from django.conf import settings

if settings.DJANGO_SSO_BACKEND_ENABLED:
    import time
    import os
    import re
    from selenium.webdriver.common.action_chains import ActionChains
    from selenium.webdriver.common.keys import Keys
    from django.core import mail
    from django.conf import settings

    from .base import TestFunzionale


    class NuovoVisitatoreTest(TestFunzionale):

        def test_puo_registrarsi_al_sito(self):
            step = 0
            # L'utente si connette all'indirizzo della pagina
            app_url = self.server_url + '?next=' + self.service.url
            self.browser.get(app_url)
            time.sleep(6)

            # riconosce la pagina
            self.identifica_titolo('Django SSO App')

            self.click_link('/login/')

            # che gli chiede di inserire le credenziali di accesso
            self.tag_contiene_testo('p', 'If you have not created an account yet, then please sign up first.')

            # prova a loggarsi
            self.compila_input_by_name('login', 'pippo@disney.com')

            self.compila_input_by_name('password', 'paperina')

            # vede che non è possibile senza creare un nuovo utente
            self.tag_contiene_testo('li', 'The e-mail address and/or password you specified are not correct.')

            # segue il link di registrazione
            self.click_link('/signup/')

            # inserisce email giusta
            self.compila_input_by_name('email', 'pippo@disney.com')

            # inserisce username
            self.compila_input_by_name('username', 'pippo')

            # inserisce password
            self.compila_input_by_name('password1', 'paperina')

            # conferma password
            self.compila_input_by_name('password2', 'paperina')

            # mette nome
            self.compila_input_by_name('first_name', 'Pippo')

            # mette cognome
            self.compila_input_by_name('last_name', 'Goofy')

            # mette description
            self.compila_input_by_name('description', 'Yuk! Yuk yuk yuk yuk!! YUK!', el="textarea")

            # inserisce immagine
            avatar_path = os.path.join(settings.ROOT_DIR, 'backend', 'tests', 'assets', 'avatar.png')
            self.browser.find_element_by_name('picture').send_keys(avatar_path)

            # inserisce posizione
            campo = self.compila_input_by_name('address', 'viale ettore andreis 74 desenzano', el="textarea")
            self.compila_input_by_name('latitude', '0')
            self.compila_input_by_name('longitude', '0')
            self.compila_input_by_name('country', 'it')

            # Lo si informa dell'avvenuta registrazione
            self.tag_contiene_testo('p', "We have sent an e-mail to you for verification. "
                                         "Follow the link provided to finalize the signup process. "
                                         "Please contact us if you do not receive it within a few minutes.")

            # Controlla la casella della posta in arrivo e segue il link di attivazione
            self.assertEqual(len(mail.outbox), 1)
            message = str(mail.outbox[0].message())
            print('MESSAGE!', message)
            activate_email_url = re.search("(?P<url>https?://.[^\s]+)", message).group("url")
            activate_email_url = activate_email_url.replace('http://' + settings.APP_DOMAIN, '')

            print('EMAIL_URL', activate_email_url, self.server_url + activate_email_url)
            self.browser.get(self.server_url + activate_email_url)

            time.sleep(1)

            # il link di attivazione lo redirige alla pagina di login
            self.tag_contiene_testo('p', 'If you have not created an account yet, then please sign up first.')

            # inserise le credenziali
            self.compila_input_by_name('login', 'pippo@disney.com')
            self.compila_input_by_name('password', 'paperina')

            time.sleep(2)

            self.tag_contiene_testo('h1', 'Welcome pippo')

            print('\n\nEOT')
            time.sleep(3)



from django.conf import settings

if settings.DJANGO_SSO_BACKEND_ENABLED:
    from django.test import TestCase
    from allauth.account.models import EmailAddress
    from django.contrib.auth import get_user_model

    from .models import Group

    User = get_user_model()


    class UserTestCase(TestCase):
        def setUp(self):
            self.user = User.objects.create(username="pippo", email="pippo@disney.com")
            self.user_email = EmailAddress.objects.create(user=self.user,
                                                          email=self.user.email,
                                                          primary=True,
                                                          verified=True)
            self.group = Group.objects.create(name='new_group')


        def test_add_user_to_group_updates_rev(self):
            user = self.user
            user_rev = user.sso_rev

            user.groups.add(self.group)

            print('groups', user.groups)

            self.assertEqual(user.sso_rev, user_rev + 1)



import random

from django.conf import settings
from django.db import models, IntegrityError


from django_sso_app.core.models import CreatedAtModel
from ..devices.models import Device
from .settings import (
  CODE_LENGTH,
  CODE_CHARS,
)


class PassepartoutManager(models.Manager):
    def create_passepartout(self, device, jwt):
        passepartout = self.create(
            device=device,
            jwt=jwt,
            token=Passepartout.generate_token()
        )
        try:
            passepartout.save()
        except IntegrityError:
            # Try again with other code
            passepartout = Passepartout.objects.create_passepartout(device, jwt)
        return passepartout


class Passepartout(CreatedAtModel):
    objects = PassepartoutManager()

    deleted = models.BooleanField(default=False)

    token = models.CharField(unique=True, db_index=True, max_length=36)
    jwt = models.TextField()
    device = models.ForeignKey(Device, verbose_name="device", on_delete=models.CASCADE)
    bumps = models.PositiveIntegerField(default=len(settings.DJANGO_SSO_DOMAINS)-1)

    @classmethod
    def generate_token(cls):
        return "".join(random.choice(CODE_CHARS) for i in range(CODE_LENGTH))

    class Meta:
        unique_together = ('token', 'deleted')

    def __str__(self):
        return '{0}:{1}'.format(self.id, self.device.fingerprint)



import logging
import os
import platform

from django.urls import reverse
from django.conf import settings
from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.schemas import SchemaGenerator
from rest_framework.views import APIView
from rest_framework_swagger import renderers

logger = logging.getLogger('backend')

CURRENT_DIR = os.getcwd()


if platform.system() == 'Windows':
    def local_space_available(dir):
        """Return space available on local filesystem."""
        import ctypes
        free_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(dir), None, None, ctypes.pointer(free_bytes))
        return free_bytes.value
else:
    def local_space_available(dir):
        destination_stats = os.statvfs(dir)
        return destination_stats.f_bsize * destination_stats.f_bavail


class StatsView(APIView):
    """
    Return instance stats
    """

    permission_classes = (AllowAny,)

    def get(self, request):
        try:
            #stats = os.statvfs(CURRENT_DIR)
            free_space_mb = int(local_space_available(CURRENT_DIR) / (1024 * 1024))
            # free_space_mb = int(
            #     (stats.f_bavail * stats.f_frsize) / (1024 * 1024))

            logger.info(
                'Free space (MB): {}.'.format(free_space_mb))

            if free_space_mb > 200:
                health_status = 'green'
            else:
                if free_space_mb < 100:
                    health_status = 'yellow'
                else:
                    health_status = 'red'

            data = {
                'status': health_status,
            }

            if request.user is not None and request.user.is_staff:
                data['free_space_mb'] = free_space_mb

            return Response(data, status.HTTP_200_OK)

        except Exception as e:
            err_msg = str(e)
            logger.exception('Error getting health {}'.format(err_msg))
            return Response(err_msg, status.HTTP_500_INTERNAL_SERVER_ERROR)


class APIRoot(APIView):
    """
    API Root
    """
    permission_classes = (AllowAny,)

    def get(self, request, *args, **kwargs):
        try:
            patterns = {
                'stats': reverse('stats', request=request),
            }
            if settings.DJANGO_SSO_BACKEND_ENABLED or settings.DJANGO_SSO_APP_ENABLED:
                patterns['auth'] = reverse('authapi', request=request, *args, **kwargs)
                patterns['profiles'] = reverse('profiles:list', request=request, *args, **kwargs)

            return Response(patterns)
        except:
            logger.exception('Error getting api-root')


class SwaggerSchemaView(APIView):
    """
    OpenAPI
    """
    permission_classes = (AllowAny,)
    renderer_classes = (
        renderers.OpenAPIRenderer,
        renderers.SwaggerUIRenderer
    )
    title = 'Django Fisherman'
    patterns = []

    def get(self, request):
        generator = SchemaGenerator(title=self.title, patterns=self.patterns)
        schema = generator.get_schema(request=request)

        return Response(schema)



import logging
import os

from django.conf import settings
from django.urls import reverse
from django.views.generic import TemplateView
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.views import APIView

logger = logging.getLogger('backend')

CURRENT_DIR = os.getcwd()


class AuthAPIRoot(APIView):
    """
    Auth API Root
    """
    permission_classes = (AllowAny,)

    def get(self, request, *args, **kwargs):
        try:
            return Response({
                'users': reverse('users:list', request=request, *args, **kwargs),
                'groups': reverse('groups:list', request=request, *args, **kwargs),

                'login': reverse('rest_login', request=request, *args, **kwargs),
                'logout': reverse('rest_logout', request=request, *args, **kwargs),
                'subscribe': reverse('rest_register', request=request, *args, **kwargs),
            })
        except:
            logger.exception('Error getting auth-api-root')


# custom views

class WebpackBuiltTemplateView(TemplateView):
    """
    Base built template view
    """
    template_name = "frontend/built.html"

    def get_context_data(self, *args, **kwargs):
        context = super(WebpackBuiltTemplateView, self).get_context_data(*args,
                                                                         **kwargs)
        context['COOKIE_DOMAIN'] = settings.COOKIE_DOMAIN
        context['DJANGO_SSO_DEFAULT_REFERRER'] = settings.DJANGO_SSO_DEFAULT_REFERRER
        return context


class ProfileView(TemplateView):
    template_name = "pages/profile.html"

    def get(self, *args, **kwargs):
        logger.info('Getting profile')
        try:
            return super(ProfileView, self).get(*args, **kwargs)
        except:
            logger.exception('Error getting profile')
            raise


class ProfileUpdateView(WebpackBuiltTemplateView):

    def get(self, *args, **kwargs):
        logger.info('Getting profile update')
        try:
            return super(ProfileUpdateView, self).get(*args, **kwargs)
        except:
            logger.exception('Error getting profile update')
            raise



import logging

from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from django_sso_app.core.models import CreatedAtModel, UpdatableModel, DeactivableModel

logger = logging.getLogger('profiles')


class AbstractUserProfileModel(CreatedAtModel, UpdatableModel, DeactivableModel):
    """
    User profile abstract model
    """

    class Meta:
        abstract = True

    sso_id = models.CharField(max_length=36, unique=True, null=True, blank=True, db_index=True)  # unique on not null
    sso_rev = models.PositiveIntegerField(default=0)

    first_name = models.CharField(_('first name'), max_length=30, null=True,
                                  blank=True)
    last_name = models.CharField(_('last name'), max_length=30, null=True,
                                 blank=True)

    role = models.SmallIntegerField(null=True, blank=True)

    description = models.TextField(_('description'), null=True, blank=True)
    picture = models.TextField(_('picture'), null=True, blank=True)
    birthdate = models.DateField(_('birthdate'), null=True, blank=True)

    latitude = models.FloatField(_('latitude'), null=True, blank=True)
    longitude = models.FloatField(_('longitude'), null=True, blank=True)

    country = models.CharField(_('country'), max_length=46, null=True,
                               blank=True)

    address = models.TextField(_('address'), null=True, blank=True)
    language = models.CharField(_('language'), null=True, blank=True)

    unsubscribed_at = models.DateTimeField(null=True, blank=True)

    language = models.CharField(max_length=3, null=True, blank=True)

    @property
    def username(self):
        return self.user.username

    @property
    def email(self):
        return self.user.email

    @property
    def is_unsubscribed(self):
        return self.unsubscribed_at is not None

    @is_unsubscribed.setter
    def is_unsubscribed(self, value):
        self.unsubscribed_at = timezone.now()

    def __str__(self):
        return self.user.username



from django.urls import path

from django_sso_app.backend.users.views import (
    user_redirect_view,
    user_update_view,
    user_detail_view,
)

app_name = "users"
urlpatterns = [
    path("~redirect/", view=user_redirect_view, name="redirect"),
    path("~update/", view=user_update_view, name="user.update"),
    path("<str:username>/", view=user_detail_view, name="user-detail"),
]



from django.views.generic import TemplateView
from django.conf import settings


class WebpackBuiltTemplateView(TemplateView):
    """
    Base built template view
    """
    template_name = "frontend/built.html"

    def get_context_data(self, *args, **kwargs):
        context = super(WebpackBuiltTemplateView, self).get_context_data(*args,
                                                                         **kwargs)
        context['COOKIE_DOMAIN'] = settings.COOKIE_DOMAIN
        context['DJANGO_SSO_DEFAULT_REFERRER'] = settings.DJANGO_SSO_DEFAULT_REFERRER
        return context



import logging

from django.conf import settings
from rest_framework import authentication

from django_sso_app.core.utils import get_profile_model
from .utils import create_local_profile_from_headers, get_sso_id

logger = logging.getLogger('app')
Profile = get_profile_model()


def get_apigw_user(request):
    is_anonymous = request.META.get(settings.DJANGO_SSO_ANONYMOUS_CONSUMER_HEADER, None) is not None
    apigw_consumer_id = get_sso_id(
        request.META.get(settings.DJANGO_SSO_CONSUMER_USERNAME_HEADER, settings.DJANGO_SSO_ANONYMOUS_CONSUMER_USERNAME))

    logger.debug('Got apigw_consumer_id: {}'.format(apigw_consumer_id))

    # Consumer is anonymous
    if is_anonymous or apigw_consumer_id == settings.DJANGO_SSO_ANONYMOUS_CONSUMER_USERNAME:
        logger.info('Apigw says user is anonymous')
        return

    try:
        user_profile = Profile.objects.get(sso_id=apigw_consumer_id)
        user = user_profile.user
        if settings.DJANGO_SSO_APP_USES_EXTERNAL_USER_PROFILE:
            user = user_profile.user

    except Profile.DoesNotExist:
        logger.info(
            'User with sso_id: {} does not exists, creating new one'.format(apigw_consumer_id))
        user = create_local_profile_from_headers(request)

    finally:
        return user


class DjangoSsoAppAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        user = get_apigw_user(request)
        logger.info('DjangoSsoAppAuthentication: authenticated as {}'.format(user))
        return user, None



# Import your application as:
# from wsgi import application
# Example:

# Import CherryPy
import os
import cherrypy

from django_sso_app.core.settings.backend import DJANGO_SSO_SERVER_PORT
from django_sso_app.backend.config.wsgi import application


def run(host="0.0.0.0", port=DJANGO_SSO_SERVER_PORT):
    os.environ.setdefault("DJANGO_SSO_APP_SERVER", "True")

    # Mount the application
    cherrypy.tree.graft(application, "/")

    # Unsubscribe the default server
    cherrypy.server.unsubscribe()

    # Instantiate a new server object
    server = cherrypy._cpserver.Server()

    # Configure the server object
    server.socket_host = host
    server.socket_port = port
    server.thread_pool = 30

    # For SSL Support
    # server.ssl_module            = 'pyopenssl'
    # server.ssl_certificate       = 'ssl/certificate.crt'
    # server.ssl_private_key       = 'ssl/private.key'
    # server.ssl_certificate_chain = 'ssl/bundle.crt'

    # Subscribe this server
    server.subscribe()

    # Example for a 2nd server (same steps as above):
    # Remember to use a different port

    # server2             = cherrypy._cpserver.Server()

    # server2.socket_host = "0.0.0.0"
    # server2.socket_port = 8081
    # server2.thread_pool = 30
    # server2.subscribe()

    # Start the server engine (Option 1 *and* 2)

    cherrypy.engine.start()
    cherrypy.engine.block()


if __name__ == '__main__':
    run()



import logging

from django.conf import settings
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import activate as activate_translation

from allauth.account import app_settings as allauth_settings
from allauth.utils import get_username_max_length
from allauth.account.models import EmailAddress
from allauth.account.adapter import get_adapter

from rest_auth.registration.serializers import RegisterSerializer as rest_authRegisterSerializer
from rest_auth.serializers import LoginSerializer as rest_authLoginSerializer
from rest_auth.serializers import PasswordResetSerializer as rest_authPasswordResetSerializer
from rest_framework import serializers, exceptions
from rest_framework.exceptions import ValidationError

from django_sso_app.core.apps.users.serializers import UserSerializer
from django_sso_app.core.utils import get_country_language

logger = logging.getLogger('backend')


class LoginSerializer(rest_authLoginSerializer):
    username = serializers.CharField(required=False, allow_null=True)
    email = serializers.EmailField(required=False, allow_null=True)
    password = serializers.CharField(style={'input_type': 'password'})
    fingerprint = serializers.CharField()

    def _validate_fingerprint(self, fingerprint):
        if fingerprint is None:
            msg = _('Must include valid "fingerprint".')
            raise exceptions.ValidationError(msg)

        return fingerprint


class RegisterSerializer(rest_authRegisterSerializer):

    username = serializers.CharField(
        max_length=get_username_max_length(),
        min_length=allauth_settings.USERNAME_MIN_LENGTH,
        required=allauth_settings.USERNAME_REQUIRED
    )
    email = serializers.EmailField(required=allauth_settings.EMAIL_REQUIRED)
    password1 = serializers.CharField(write_only=True)
    password2 = serializers.CharField(write_only=True)

    first_name = serializers.CharField(required=True)
    last_name = serializers.CharField(required=False, default='')

    description = serializers.CharField(required=False, allow_null=True, allow_blank=True)
    picture = serializers.CharField(required=False, allow_null=True)
    birthdate = serializers.DateField(required=False, allow_null=True)

    latitude = serializers.FloatField(required=True)
    longitude = serializers.FloatField(required=True)

    country = serializers.CharField(required=True)
    address = serializers.CharField(required=False, allow_null=True)

    language = serializers.CharField(required=False, allow_null=True)

    role = serializers.IntegerField(required=False)

    referrer = serializers.CharField(required=True)

    def get_cleaned_data(self):
        d = {
            'email': self.validated_data.get('email', None),
            'username': self.validated_data.get('username', None),
            'password1': self.validated_data.get('password1', self.validated_data.get('password', None)), # adds password check for staff creations
            'password2': self.validated_data.get('password2', None),
            'referrer': self.validated_data.get('referrer', None)
        }

        for f in settings.DJANGO_SSO_REQUIRED_PROFILE_FIELDS:
            d[f] = self.validated_data.get(f, None)

        if d['language'] is None:
            d['language'] = get_country_language(d['country'])

        return d

    def custom_signup(self, request, user):
        adapter = get_adapter(request)
        referrer = self.cleaned_data.get('referrer')
        adapter.subscribe_to_service(user, referrer)

    def save(self, request):
        try:
            return super(RegisterSerializer, self).save(request)

        except Exception as e:
            logger.error('Error registering user: {}'.format(e), exc_info=True)
            raise


class PasswordResetSerializer(rest_authPasswordResetSerializer):
    """
    Uses allauth default email templates
    """
    def validate_email(self, value):
        value = super(PasswordResetSerializer, self).validate_email(value)
        saved_email = EmailAddress.objects.get(email=value)
        if not saved_email.verified:
            raise ValidationError(
                _('E-mail address not verified: %(value)s'),
                code='invalid',
                params={'value': value},
            )

        user = saved_email.user
        user_language = user.language
        logger.info('Rendering mail for {0} with language {1}'.format(user, user_language))
        activate_translation(user_language)

        return value

    def get_email_options(self):
        """Override this method to change default e-mail options"""

        return {
            'email_template_name': 'account/email/password_reset_key_message.txt',
            'html_email_template_name': 'account/email/password_reset_key_message.html',
            'use_https': settings.ENABLE_HTTPS,
            'extra_email_context': {
                'EMAILS_DOMAIN': settings.EMAILS_DOMAIN,
                'EMAILS_SITE_NAME': settings.EMAILS_SITE_NAME
            }
        }

class CsrfTokenSerializer(serializers.Serializer):
    token = serializers.CharField()


class JWTSerializer(serializers.Serializer):
    """
    Serializer for JWT authentication.
    """
    token = serializers.CharField()
    user = UserSerializer()
    passepartout_redirect_url = serializers.CharField(required=False)



LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '%(levelname)s:%(name)s: %(message)s '
                      '(%(asctime)s; %(filename)s:%(lineno)d)',
            'datefmt': "%Y-%m-%d %H:%M:%S",
        },
    },
    'handlers': {
        'console':{
            'level':'DEBUG',
            'class':'logging.StreamHandler',
            'formatter': 'simple'
        },
        'mail_admins': {
            'level': 'ERROR',
            'class': 'django.utils.log.AdminEmailHandler',
        },
    },
    'loggers': {
        'django': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'users': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'profiles': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'devices': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'services': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'passepartout': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'api_gateway': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'backend': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'app': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True,
        },
        'django_sso_app': {
            'handlers': ['console'],
            'propagate': True,
            'level': 'DEBUG'
        },
        'django': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
    },
}
